//============================================================================
// Name        : ShapeFactory.cpp
// Author      : Vikturek
// Version     :
// Copyright   : Your copyright notice
// Description : Hello World in C++, Ansi-style
//============================================================================

#include <iostream>
#include <fstream>
#include <map>
#include <vector>

using namespace std;

#define DRAWING_FILE "Drawing.txt"

class Shape {

public:
	virtual void draw() = 0;
	virtual void load(ifstream &file) = 0;
	virtual ~Shape(){}
};

typedef Shape* (*ShapeCreateFunc)();
typedef map<int, ShapeCreateFunc> ShapesRegister;

class ShapeFactory {

	ShapesRegister shapesRegister;
	static ShapeFactory* factory;
public:
	static ShapeFactory* Instance(){
		if(factory == NULL){
			factory = new ShapeFactory;
		}
		return factory;
	}
	int registerShape(int shapeId, ShapeCreateFunc callback){
		shapesRegister[shapeId] = callback;
		return 0;
	}
	void unregisterShape(int shapeId){
		ShapesRegister::iterator it = shapesRegister.find(shapeId);
		if(it != shapesRegister.end()){
			shapesRegister.erase(it);
		}
	}
	Shape *createShape(int shapeId){
		return shapesRegister[shapeId]();
	}
};

ShapeFactory *ShapeFactory::factory = NULL;

class ShapeCircle: public Shape {
#define SHAPE_CIRCLE 1
	double x;
	double y;
	double r;
public:
	ShapeCircle(){}
	ShapeCircle(double _x, double _y, double _r):
		x(_x), y(_y), r(_r){}
	virtual void draw(){
		cout << "Circle: " << x << " " << y << " " << r << endl;
 	}
	virtual void load(ifstream &file){
		file >> x;
		file >> y;
		file >> r;
	}
	virtual ~ShapeCircle(){}
};

namespace {
	Shape *createCircle(){
		return new ShapeCircle;
	}

	int a = ShapeFactory::Instance()->registerShape(SHAPE_CIRCLE, createCircle);
};

class ShapeSquare: public Shape {
#define SHAPE_SQUARE 2
	double x;
	double y;
	double a;
public:
	ShapeSquare(){}
	ShapeSquare(double _x, double _y, double _a):
		x(_x), y(_y), a(_a){}
	virtual void draw(){
		cout << "Square: " << x << " " << y << " " << a << endl;
	}
	virtual void load(ifstream &file){
		file >> x;
		file >> y;
		file >> a;
	}
	virtual ~ShapeSquare(){}
};

namespace {
	Shape *createSquare(){
		return new ShapeSquare;
	}

	int b = ShapeFactory::Instance()->registerShape(SHAPE_SQUARE, createSquare);
};

class Drawing {
public:
	typedef vector<Shape *> Shapes;
private:
	Shapes shapes;
public:
	void addShape(Shape *shape){
		shapes.push_back(shape);
	}
	void draw(){
		for(Shapes::iterator i = shapes.begin(); i != shapes.end(); ++i){
			(*i)->draw();
		}
	}
	void load(const char *fileName){
		ifstream file(fileName);
		int shapeId;
		Shape *shape;
		file >> shapeId;
		while(not file.eof()){
			shape = ShapeFactory::Instance()->createShape(shapeId);
			shape->load(file);
			shapes.push_back(shape);
			file >> shapeId;
		}
	}
};

int main() {
	Drawing drawing;

	drawing.addShape(new ShapeCircle(1, 1, 2));
	drawing.addShape(new ShapeSquare(2, 2, 4));
	drawing.draw();
	drawing.load(DRAWING_FILE);
	drawing.draw();
	return 0;
}

